@@ -27,24 +27,42 @@ class LdapAuthConfig
         return $this->requireString('ldapUri');
     }
 
+    public function tlsCa(): ?string
+    {
+        return $this->optionalString('tlsCa');
+    }
+
+    public function tlsCert(): ?string
+    {
+        return $this->optionalString('tlsCert');
+    }
+
+    public function tlsKey(): ?string
+    {
+        return $this->optionalString('tlsKey');
+    }
+
     public function bindDnTemplate(): ?string
     {
         return $this->optionalString('bindDnTemplate');
     }
 
-    public function baseDn(): ?string
+    /**
+     * @return array<string>
+     */
+    public function baseDn(): array
     {
-        return $this->optionalString('baseDn');
+        return $this->requireStringOrStringArray('baseDn');
     }
 
-    public function userFilterTemplate(): ?string
+    public function userFilterTemplate(): string
     {
-        return $this->optionalString('userFilterTemplate');
+        return $this->requireString('userFilterTemplate');
     }
 
-    public function userIdAttribute(): ?string
+    public function userIdAttribute(): string
     {
-        return $this->optionalString('userIdAttribute');
+        return $this->requireString('userIdAttribute');
     }
 
     public function addRealm(): ?string
