@@ -11,6 +11,7 @@ declare(strict_types=1);
 
 namespace Vpn\Portal;
 
+use DateInterval;
 use DateTimeImmutable;
 use Vpn\Portal\Cfg\Config;
 use Vpn\Portal\Cfg\ProfileConfig;
@@ -40,6 +41,23 @@ class ConnectionManager
         $this->dateTime = Dt::get();
     }
 
+    /**
+     * @return array<array{node_number:int,node_url:string,node_info:?array{rel_load_average:array<int>,load_average:array<float>,cpu_count:int,node_uptime:int}}>
+     */
+    public function nodeInfo(): array
+    {
+        $nodeInfoList = [];
+        foreach ($this->config->nodeNumberUrlList() as $nodeNumber => $nodeUrl) {
+            $nodeInfoList[] = [
+                'node_number' => $nodeNumber,
+                'node_url' => $nodeUrl,
+                'node_info' => $this->vpnDaemon->nodeInfo($nodeUrl),
+            ];
+        }
+
+        return $nodeInfoList;
+    }
+
     /**
      * Retrieve a list of all current OpenVPN and WireGuard connections, sorted
      * by profile.
@@ -119,7 +137,6 @@ class ConnectionManager
     {
         foreach ($this->storage->oCertInfoListByAuthKey($authKey) as $oCertInfo) {
             $this->oDisconnect(
-                $oCertInfo['user_id'],
                 $oCertInfo['profile_id'],
                 $oCertInfo['node_number'],
                 $oCertInfo['common_name']
@@ -142,7 +159,6 @@ class ConnectionManager
     {
         foreach ($this->storage->oCertInfoListByUserId($userId) as $oCertInfo) {
             $this->oDisconnect(
-                $userId,
                 $oCertInfo['profile_id'],
                 $oCertInfo['node_number'],
                 $oCertInfo['common_name']
@@ -187,7 +203,6 @@ class ConnectionManager
                 return;
             }
             $this->oDisconnect(
-                $oCertInfo['user_id'],
                 $oCertInfo['profile_id'],
                 $oCertInfo['node_number'],
                 $connectionId
@@ -204,15 +219,29 @@ class ConnectionManager
      */
     public function connect(ServerInfo $serverInfo, ProfileConfig $profileConfig, string $userId, array $clientProtoSupport, string $displayName, DateTimeImmutable $expiresAt, bool $preferTcp, ?string $publicKey, ?string $authKey): ClientConfigInterface
     {
-        if ('wireguard' === Protocol::determine($profileConfig, $clientProtoSupport, $publicKey, $preferTcp)) {
-            if (null !== $publicKey) {
-                if (null !== $wireGuardConfig = $this->wConnect($serverInfo, $profileConfig, $userId, $displayName, $expiresAt, $publicKey, $authKey)) {
+        foreach (Protocol::determine($profileConfig, $clientProtoSupport, $publicKey, $preferTcp) as $vpnProto) {
+            switch ($vpnProto) {
+                case 'wireguard':
+                    /**
+                     * Protocol::determine makes sure $publicKey is NOT null.
+                     *
+                     * @var string $publicKey
+                     */
+                    if (null === $wireGuardConfig = $this->wConnect($serverInfo, $profileConfig, $userId, $displayName, $expiresAt, $publicKey, $authKey)) {
+                        break;
+                    }
+
                     return $wireGuardConfig;
-                }
+                case 'openvpn':
+                    if (null === $openVpnConfig = $this->oConnect($serverInfo, $profileConfig, $userId, $displayName, $expiresAt, $preferTcp, $authKey)) {
+                        break;
+                    }
+
+                    return $openVpnConfig;
             }
         }
 
-        return $this->oConnect($serverInfo, $profileConfig, $userId, $displayName, $expiresAt, $preferTcp, $authKey);
+        throw new ConnectionManagerException('unable to connect using any of the common supported VPN protocols');
     }
 
     /**
@@ -248,7 +277,7 @@ class ConnectionManager
      * (3) Profile MUST still exist
      * (4) Node MUST still exist
      *
-     * @return array<string,array{node_number:int,user_id:string,profile_id:string,display_name:string,common_name:string,expires_at:\DateTimeImmutable,auth_key:?string,user_is_disabled:bool}>
+     * @return array<string, array{auth_key: null|string, common_name: string, created_at: DateTimeImmutable, display_name: string, expires_at: DateTimeImmutable, node_number: int, profile_id: string, user_id: string}>
      */
     private function oFilterDb(): array
     {
@@ -258,6 +287,7 @@ class ConnectionManager
             if ($oCertInfo['user_is_disabled']) {
                 continue;
             }
+            unset($oCertInfo['user_is_disabled']);
             if ($oCertInfo['expires_at'] <= $this->dateTime) {
                 continue;
             }
@@ -267,7 +297,7 @@ class ConnectionManager
             }
             $profileConfig = $this->config->profileConfig($profileId);
             $nodeNumber = $oCertInfo['node_number'];
-            if (!in_array($nodeNumber, $profileConfig->onNode())) {
+            if (!in_array($nodeNumber, $profileConfig->onNode(), true)) {
                 continue;
             }
             $oFilteredCertList[$commonName] = $oCertInfo;
@@ -287,7 +317,7 @@ class ConnectionManager
      * (5) IPv4 address MUST belong to prefix of Profile (+Node)
      * (6) IPv6 address MUST belong to prefix of Profile (+Node)
      *
-     * @return array<string,array{node_number:int,user_id:string,profile_id:string,display_name:string,public_key:string,ip_four:string,ip_six:string,expires_at:\DateTimeImmutable,auth_key:?string,user_is_disabled:bool}>
+     * @return array<string, array{auth_key: null|string, created_at: DateTimeImmutable, display_name: string, expires_at: DateTimeImmutable, ip_four: string, ip_six: string, node_number: int, profile_id: string, public_key: string, user_id: string}>
      */
     private function wFilterDb(): array
     {
@@ -297,6 +327,7 @@ class ConnectionManager
             if ($wPeerInfo['user_is_disabled']) {
                 continue;
             }
+            unset($wPeerInfo['user_is_disabled']);
             if ($wPeerInfo['expires_at'] <= $this->dateTime) {
                 continue;
             }
@@ -306,7 +337,7 @@ class ConnectionManager
             }
             $profileConfig = $this->config->profileConfig($profileId);
             $nodeNumber = $wPeerInfo['node_number'];
-            if (!in_array($nodeNumber, $profileConfig->onNode())) {
+            if (!in_array($nodeNumber, $profileConfig->onNode(), true)) {
                 continue;
             }
             $ipFour = Ip::fromIp($wPeerInfo['ip_four']);
@@ -339,19 +370,52 @@ class ConnectionManager
 
     private function wSync(): void
     {
+        $appGoneInterval = $this->config->apiConfig()->appGoneInterval();
+        $appGoneCutoffMoment = $this->dateTime->sub($appGoneInterval);
+
         $wPeerListFromDb = $this->wFilterDb();
         $connectedPublicKeyList = [];
         foreach ($this->config->nodeNumberUrlList() as $nodeNumber => $nodeUrl) {
-            foreach (array_keys($this->vpnDaemon->wPeerList($nodeUrl, true)) as $publicKey) {
+            $nodeUptime = new DateInterval('PT0S');
+            if (null !== $nodeInfo = $this->vpnDaemon->nodeInfo($nodeUrl)) {
+                $nodeUptime = new DateInterval('PT'.$nodeInfo['node_uptime'].'S');
+            }
+            // we require the node to be up at least "appGoneInterval" before
+            // checking whether API clients are "gone"
+            $nodeHasAdequateUptime = $this->dateTime->sub($nodeUptime)->add($appGoneInterval) < $this->dateTime;
+            foreach ($this->vpnDaemon->wPeerList($nodeUrl, true) as $publicKey => $wPeerInfo) {
                 if (!\array_key_exists($publicKey, $wPeerListFromDb)) {
-                    if (null !== $openConnectionInfo = $this->storage->openConnectionInfo($publicKey)) {
-                        $this->logger->debug(sprintf('%s: removing peer [%s,%d,%s]', __METHOD__, $openConnectionInfo['profile_id'], $nodeNumber, $publicKey));
-                        $this->wDisconnect($openConnectionInfo['user_id'], $openConnectionInfo['profile_id'], $nodeNumber, $publicKey, $openConnectionInfo['ip_four'], $openConnectionInfo['ip_six']);
-                    }
-                    // XXX should we log when there is no open connection for this public key in the connection_log table?
+                    $this->wSyncDisconnect($nodeNumber, $publicKey, 'public key not in database');
 
                     continue;
                 }
+                if ($nodeHasAdequateUptime) {
+                    $wPeerDbInfo = $wPeerListFromDb[$publicKey];
+                    if (null !== $wPeerDbInfo['auth_key']) {
+                        // WireGuard configuration was issued to an API client
+                        if ($wPeerDbInfo['created_at'] < $appGoneCutoffMoment) {
+                            // configuration must have been created before the
+                            // cutoff
+                            if (null === $lastHandshakeTime = $wPeerInfo['last_handshake_time']) {
+                                $this->wSyncDisconnect($nodeNumber, $publicKey, 'no handshake since node boot');
+                                // make sure it does not get added again in
+                                // next section
+                                unset($wPeerListFromDb[$publicKey]);
+
+                                continue;
+                            }
+                            if (new DateTimeImmutable($lastHandshakeTime) < $appGoneCutoffMoment) {
+                                $this->wSyncDisconnect($nodeNumber, $publicKey, 'handshake too long ago');
+                                // make sure it does not get added again in
+                                // next section
+                                unset($wPeerListFromDb[$publicKey]);
+
+                                continue;
+                            }
+                        }
+                    }
+                }
+
                 $connectedPublicKeyList[] = $publicKey;
             }
         }
@@ -370,6 +434,20 @@ class ConnectionManager
         }
     }
 
+    /**
+     * Remove/disconnect a peer during "sync" stage based on various criteria.
+     *
+     * @see ConnectionManager::wSync()
+     */
+    private function wSyncDisconnect(int $nodeNumber, string $publicKey, string $disconnectReason): void
+    {
+        if (null !== $openConnectionInfo = $this->storage->openConnectionInfo($publicKey)) {
+            $this->logger->debug(sprintf('%s: removing peer (reason: %s) [%s,%d,%s]', __METHOD__, $disconnectReason, $openConnectionInfo['profile_id'], $nodeNumber, $publicKey));
+            $this->wDisconnect($openConnectionInfo['user_id'], $openConnectionInfo['profile_id'], $nodeNumber, $publicKey, $openConnectionInfo['ip_four'], $openConnectionInfo['ip_six']);
+        }
+        // XXX should we log when there is no open connection for this public key in the connection_log table?
+    }
+
     private function wDisconnect(string $userId, string $profileId, int $nodeNumber, string $publicKey, string $ipFour, string $ipSix): void
     {
         if (null === $nodeUrl = $this->nodeUrl($nodeNumber)) {
@@ -389,7 +467,7 @@ class ConnectionManager
         $this->connectionHook->disconnect($userId, $profileId, 'wireguard', $publicKey, $ipFour, $ipSix, $bytesIn, $bytesOut);
     }
 
-    private function oDisconnect(string $userId, string $profileId, int $nodeNumber, string $commonName): void
+    private function oDisconnect(string $profileId, int $nodeNumber, string $commonName): void
     {
         if (null === $nodeUrl = $this->nodeUrl($nodeNumber)) {
             $this->logger->warning(sprintf('node "%d" does not exist (anymore)', $nodeNumber));
@@ -438,7 +516,7 @@ class ConnectionManager
                 $ipFour,
                 $ipSix,
                 $serverPublicKey,
-                $this->config->wireGuardConfig()->listenPort(),
+                $this->config->wireGuardConfig(),
                 $expiresAt
             );
         }
@@ -449,7 +527,7 @@ class ConnectionManager
         return null;
     }
 
-    private function oConnect(ServerInfo $serverInfo, ProfileConfig $profileConfig, string $userId, string $displayName, DateTimeImmutable $expiresAt, bool $preferTcp, ?string $authKey): OpenVpnClientConfig
+    private function oConnect(ServerInfo $serverInfo, ProfileConfig $profileConfig, string $userId, string $displayName, DateTimeImmutable $expiresAt, bool $preferTcp, ?string $authKey): ?OpenVpnClientConfig
     {
         $nodeNumberList = $profileConfig->onNode();
         shuffle($nodeNumberList);
@@ -488,8 +566,10 @@ class ConnectionManager
             );
         }
 
-        // we found no node that is up
-        throw new ConnectionManagerException('no VPN node available');
+        // we found no suitable node to connect to...
+        $this->logger->error(sprintf('unable to find a suitable node to connect to for profile "%s"', $profileConfig->profileId()));
+
+        return null;
     }
 
     /**
